(* Some examples:
 
     Depth First Search
     Evaluate DAG
     Graph Reverese 
  
   The functional implementation consists of a functor for
   each problem. The functor has the functional graph structure
   as a parameter to facilitate future tests with alternative 
   graph implementations. 
   Each functor provides the function implemented by using
   the context function (realizing active pattern matching),
   e.g. dfs, and a function that uses the corresponding 
   predefined fold operation, e.g. dfsPD.

   The imperative implementations are contained in the structure 
   Imp, and a "functionalized" implementation obtained
   by replacing imperative arrays with balanced search trees
   is given in the structure SimpleFun.
*)


functor FunDFS (structure G:GRAPH) =
struct

  datatype tree = Branch of G.node * tree list

  fun gfold f d b u l g =
      let fun gfold1 (v,g) =
              let val (c as (_,_,lab,_),g1) = G.context (v,g)
                  val (r,g2) = gfoldn (f c,g1)
               in
                  (d (lab,r),g2)
               end
          and gfoldn ([],g)   = (u,g)
           |  gfoldn (v::l,g) =
              let val (x,g1) = gfold1 (v,g)
                  val (y,g2) = gfoldn (l,g1)
               in
                  (b (x,y),g2)
              end
              handle Match => gfoldn (l,g)
        in
           #1 (gfoldn (l,g))
       end
       
  fun fwd (_,_,_,s) = s
  fun dfs g = gfold fwd Branch (op ::) [] (List.tabulate (G.noNodes g,id)) g
  fun dfsPD g = G.gfold fwd Branch (op ::) [] (List.tabulate (G.noNodes g,id)) g
  
  val el2gr = G.el2gr
end
;


functor FunEvalDAG (structure G:GRAPH) =
struct

val & = G.&
infix &

fun drop1 (SOME x) (y::l) = if y=x then l else y::drop1 (SOME x) l
 |  drop1 x l = l
fun del v g' = let val (_,g) = G.context (v,g') in g end

datatype ('a,'b) hybrid = SRC of 'a | DEST of 'b

fun mfold1 d b u (z,v) g0 =
    let val ((pred,_,lab,suc),g) = G.context (v,g0)
     in
        case lab of
             DEST w => (w,(drop1 z pred,v,DEST w,[]) & g)
                       (* return the alread computed value w
                          and reinsert v without predecessor z *)
           |  SRC w =>
              let val (r,g2) = mfoldn d b u (SOME v,suc)
                               (([],v,SRC w,[]) & g)
                       (* insert v before the recursive call to
                          mfoldn (without edges) to enable
                          cyclic reentry *)
                  val new    = d (w,r)
               in
                  (new,(drop1 z pred,v,DEST new,[]) & del v g2)
              end
    end
and mfoldn d b u (_,[]) g = (u,g)
 |  mfoldn d b u (z,v::l) g =
    let val (x,g1) = mfold1 d b u (z,v) g
        val (y,g2) = mfoldn d b u (z,l) g1
     in
        (b (x,y),g2)
    end
(*    handle Match => mfoldn d b u l g      *)

fun mfold d b u l g = #1 (mfoldn d b u (NONE,l) (G.gmap SRC g))


datatype intexpr = CON of int | OP of int * int -> int

val cons = (op ::)
val append = (op @)

fun pred v g = let val ((p,_,_,_),_) = G.context (v,g) in p end
fun roots g = [G.noNodes g-1]
fun evalnode (CON i,_) = i
 |  evalnode (OP f,[x,y]) = f (x,y)
fun evaldag g = mfold evalnode cons [] (roots g) g
fun evaldagPD g = G.mfold evalnode cons [] (roots g) g

local
  fun layer (i,s,b) =
      if i<=s+b then (i-1,i-b-1)::(i-1,i-b-2)::layer (i+1,s,b) else
      if b>0 then layer (i,s+b,b-1) else []
in
   fun mkDag  n = G.mkgr (List.tabulate ((n*(n+1)) div 2,
                          fn i=> if i<n then CON i else OP Int.max))
                        (layer (n+1,n,n-1))
end

end
;


functor FunGrev (structure G:GRAPH) =
struct

fun ufold f u g =
    let val (c,g') = G.matchany g
     in f (c,ufold f u g') end
    handle Match => u

(*
fun gmap n f = ufold (fn ((p,v,l,s),g)=>G.&((p,v,f l,s),g)) (G.empty n)
*)
fun grev n = ufold (fn ((p,v,l,s),g)=>G.&((s,v,l,p),g)) (G.empty n)
fun grevPD n = G.ufold (fn ((p,v,l,s),g)=>G.&((s,v,l,p),g)) (G.empty n)

end
;


structure Imp =
struct
  datatype tree = Branch of int * tree list

  fun el2gr x el = 
      let fun tailMaxNode ([],m)       = m
           |  tailMaxNode ((v,w)::l,m) = 
              tailMaxNode (l,Int.max (Int.max (v,w),m))
          fun maxNode l = tailMaxNode (l,0)
          val n = maxNode el
          val N = Array.array (n+1,x)
          val E = Array.array (n+1,[]:int list)
          fun scan []          = ()
           |  scan ((v,w)::el) = (Array.update (E,v,w::Array.sub (E,v)); scan el)
       in
          (scan el; (N,E))
      end
  fun mkgr nl el =
      let fun tailMaxNode ([],m)       = m
           |  tailMaxNode ((v,w)::l,m) =
              tailMaxNode (l,Int.max (Int.max (v,w),m))
          fun maxNode l = tailMaxNode (l,0)
          val n = maxNode el
          val N = Array.fromList nl
          val E = Array.array (n+1,[]:int list)
          fun scan []          = ()
           |  scan ((v,w)::el) = (Array.update (E,v,w::Array.sub (E,v)); scan el
)
       in
          (scan el; (N,E))
      end


  fun dfs' d b u (N,Suc) =
      let val V = Array.array (Array.length N,false)
          fun dfs1 v = (Array.update (V,v,true);
                        d (Array.sub (N,v),dfsn (Array.sub (Suc,v))))
          and dfsn []     = u
           |  dfsn (v::l) = if Array.sub (V,v) then dfsn l
                                               else b (dfs1 v,dfsn l)
       in
          dfsn (List.tabulate (Array.length N-1,id))
      end  

  fun dfs g = dfs' Branch (op ::) [] g 


  datatype intexpr = CON of int | OP of int * int -> int

  fun roots (N,Suc) = [Array.length N-1]
  fun evaldag0 (N,Suc) =
      let fun eval1 v = 
              case Array.sub (N,v) of
                   CON i => i
                 | OP  f => f ((fn [x,y] => (x,y)) (evaln (Array.sub (Suc,v))))
          and evaln [] = []
           |  evaln (x::l) = eval1 x::evaln l
       in
          evaln (roots (N,Suc))
      end

  fun evaldag (N,Suc) =
      let val R = Array.array (Array.length N,Array.sub (N,0))
	  val _ = Array.copy {src=N,si=0,len=NONE,dst=R,di=0};
	  fun eval1 v =
              case Array.sub (R,v) of
                   CON i => i
                 | OP  f => let val result = f ((fn [x,y] => (x,y)) 
                                               (evaln (Array.sub (Suc,v))))
                             in (Array.update (R,v,CON result);result) end
          and evaln [] = []
           |  evaln (x::l) = eval1 x::evaln l
       in
          evaln (roots (N,Suc))
      end

local
  fun layer (i,s,b) =
      if i<=s+b then (i-1,i-b-1)::(i-1,i-b-2)::layer (i+1,s,b) else
      if b>0 then layer (i,s+b,b-1) else []
in
   fun mkDag  n = mkgr (List.tabulate ((n*(n+1)) div 2,
                        fn i=> if i<n then CON i else OP Int.max))
                       (layer (n+1,n,n-1))
end

  fun grev (N,Suc) =
      let val R = Array.array (Array.length N,[]:int list)
          fun scan (i,[]) = ()
           |  scan (i,v::l) = (Array.update (R,v,i::Array.sub (R,v));
                               scan (i,l))
          val _ = Array.appi scan (Suc,0,NONE)
       in 
          (N,R)
      end

end
;


structure SimpleFun =
struct
  datatype tree = Branch of int * tree list

  fun el2gr x el = 
      let val (N,E) = Imp.el2gr x el
	  fun insmap (i,a,m) = IntBinaryMap.insert (m,i,a)
       in
	  (Array.foldli insmap IntBinaryMap.empty (N,0,NONE),
	   Array.foldli insmap IntBinaryMap.empty (E,0,NONE))
      end

  fun mkgr nl el =
      let val (N,E) = Imp.mkgr nl el
	  fun insmap (i,a,m) = IntBinaryMap.insert (m,i,a)
       in
	  (Array.foldli insmap IntBinaryMap.empty (N,0,NONE),
	   Array.foldli insmap IntBinaryMap.empty (E,0,NONE))
      end

  fun forceOpt (SOME x) = x

  fun dfs' d b u (N,Suc) =
      let fun dfs1 (v,V) = 
              let val (W,w) = IntBinaryMap.remove (V,v)
		  val (x,X) = dfsn (forceOpt (IntBinaryMap.find (Suc,v)),W)
               in
                  (d (w,x),X)
              end
          and dfsn ([],V)   = (u,V)
           |  dfsn (v::l,V) = (let val (w,W) = dfs1 (v,V)
				   val (x,X) = dfsn (l,W)
                               in
                                  (b (w,x),X)
                              end handle NotFound => dfsn (l,V))
       in
          dfsn (List.tabulate (IntBinaryMap.numItems N-1,id),N)
      end  

  fun dfs g = #1 (dfs' Branch (op ::) [] g)

  fun spg (V,E) = (IntBinaryMap.listItems V,
		   IntBinaryMap.listItems E)

  datatype intexpr = CON of int | OP of int * int -> int

  fun roots (N,Suc) = [IntBinaryMap.numItems N-1]
  fun evaldag (N,Suc) =
	  let fun eval1 (v,C) = 
	      case IntBinaryMap.find (C,v) of
            SOME x => (x,C)
          | NONE =>
            (case forceOpt (IntBinaryMap.find (N,v)) of
               CON i => (i,C)
             | OP  f => 
               let val (l,C') = evaln (getOpt (IntBinaryMap.find (Suc,v),[]),C)
                   val r = f ((fn [x,y] => (x,y)) l)
                in
                   (r,IntBinaryMap.insert (C',v,r))
               end)
          and evaln ([],C)   = ([],C)
           |  evaln (x::l,C) = 
              let val (v,C')   = eval1 (x,C)
                  val (l',C'') = evaln (l,C')
               in
                  (v::l',C'')
              end
       in
          evaln (roots (N,Suc),IntBinaryMap.empty)
      end

local
  fun layer (i,s,b) =
      if i<=s+b then (i-1,i-b-1)::(i-1,i-b-2)::layer (i+1,s,b) else
      if b>0 then layer (i,s+b,b-1) else []
in
   fun mkDag  n = mkgr (List.tabulate ((n*(n+1)) div 2,
                        fn i=> if i<n then CON i else OP Int.max))
                       (layer (n+1,n,n-1))
end

  fun grev (N,Suc) =
      let val n = IntBinaryMap.numItems N
          fun scan (i,[],m)   = if i>=n then m else 
	           scan (i+1,getOpt (IntBinaryMap.find (Suc,i+1),[]),m)
           |  scan (i,v::l,m) =
	      let val (m',l') = IntBinaryMap.remove (m,v)
				handle NotFound => (m,[])
	       in
	          scan (i,l,IntBinaryMap.insert (m',v,i::l'))
	      end
       in 
          (N,scan (0,getOpt (IntBinaryMap.find (Suc,0),[]),IntBinaryMap.empty))
      end
end
;


(* generate implementations with functional graphs *)

structure FunGraph = Graph (FunArray)

structure DFS     = FunDFS (structure G=FunGraph)
structure EvalDAG = FunEvalDAG (structure G=FunGraph)
structure Grev    = FunGrev (structure G=FunGraph)


(* generate example graphs *)
  
local
  fun tailMkSuc (n,d,i,[],el)   = rev el
   |  tailMkSuc (n,d,i,x::l,el) = 
      if i>d then tailMkSuc (n,d,1,l,el) 
             else tailMkSuc (n,d,i+1,x::l,(x,(x+i) mod n)::el)      
in
   fun mkSuc (n,d,i,l) = tailMkSuc (n,d,i,l,[])
end  


(*

  (* generate some edge lists *)
  val el1K = mkSuc (1000,4,1,List.tabulate (999,id));
  val el5K = mkSuc (5000,4,1,List.tabulate (4999,id));
  val el10K = mkSuc (10000,4,1,List.tabulate (9999,id));

  (* generate example graphs *)
  val g = DFS.el2gr 7 el5K;
  val gi = Imp.el2gr 7 el5K;

  (* try algorithms *)
  DFS.dfsPD g;
  DFS.dfs g;
  Imp.dfs gi;

*)

