  (* a graph is represented by two arrays of adjacency lists storing 
     successors and predecessors. This slows down update operations
     a bit, but speeds up (full) context determination since 
     predecessors are immediately available. 
  *)


functor Graph (FunArray:FUN_ARRAY) : GRAPH =
struct

  type node = int
  type 'a context = node list * node * 'a * node list
  exception Node
  exception Edge

  open MyUtil

  (* additional array functions *)
  open FunArray
  fun apply (a,i,f)    = update (a,i,f (sub (a,i)))
  fun firstIndex (a,p) =
      let fun scan (i,p) = if p (sub (a,i)) then i else scan (i+1,p)
       in scan (0,p) end
  fun arrayToList f a =
      let fun list (f,a,i,n) = 
              if i<n then f (sub (a,i))::list (f,a,i+1,n) 
                     else []
       in list (f,a,0,size a) end
  fun impArrayToList a =
      let fun list (a,i,n) = 
              if i<n then Array.sub (a,i)::list (a,i+1,n) 
                     else []
       in list (a,0,Array.length a) end
  
  (* stamp type and operations:
     Stamps are integers whose absolute value is strictly increasing. 
     Zero or negative values represent deleted nodes.

     stampTrue   gets next stamp and makes it positive
     stampFalse  gets next stamp and makes it negative
     getStamp    simply obtains a node's stamp
     getNegStamp gets a node's negative stamp, ie, the calling
                 function insists that the node is not existing.
     getPosStamp gets a node's positive stamp, ie, the calling
                 function insists of the node's existence.
  *)
  type stamp = int
  fun stampTrue  i = abs i+1
  fun stampFalse i = ~(abs i+1)
  fun getStamp (na,n)    = sub (na,n)
  fun getNegStamp (na,n) = let val s=sub (na,n)
                            in if s>0 then raise Node else s end
  fun getPosStamp (na,n) = let val s=sub (na,n) 
                            in if s<=0 then raise Edge else s end

  datatype 'a graph = Empty of int
                    | Full  of    stamp array * 'a array
                               * (node * stamp) list array   (* pred *)
                               * (node * stamp) list array   (* suc  *)

  (* exported functions *)
  fun empty n = Empty n
  
  infixr 5 &
  fun (pred,n,l,suc)&(Empty i) = 
      (pred,n,l,suc)&(Full (array (i,0),array (i,l),array (i,[]),array (i,[])))
   |  (pred,n,l,suc)&(Full (na,la,pa,sa)) = 
      let val stampN      = stampTrue (getNegStamp (na,n))
          val stampedPred = map (fn x=>(x,getPosStamp (na,x))) pred
          val stampedSuc  = map (fn x=>(x,getPosStamp (na,x))) suc
          fun updAdj (a,[])   = a
           |  updAdj (a,v::l) = updAdj (apply (a,v,fn adj=>(n,stampN)::adj),l)
       in
          Full (update (na,n,stampN),
                update (la,n,l),
                updAdj (update (pa,n,stampedPred),suc),
                updAdj (update (sa,n,stampedSuc),pred))
      end

  fun context (n,Empty _) = raise Match
   |  context (n,Full (na,la,pred,suc)) = 
      if getStamp (na,n)>0 then
         ((select p1 (fn (v,i)=>i=getStamp (na,v)) (sub (pred,n)),n,sub (la,n),
           select p1 (fn (v,i)=>i=getStamp (na,v)) (sub (suc,n))),
          Full (apply (na,n,stampFalse),la,pred,suc))
      else
         raise Match

  fun matchany (Empty _) = raise Match
   |  matchany (g as (Full (na,la,pred,suc))) = 
               (context (firstIndex (na,fn i=>i>0),g)
                handle Subscript => raise Match)
       
   fun noNodes (Empty _)         = 0
    |  noNodes (Full (na,_,_,_)) = size na

   fun gmap f (Empty i) = Empty i
    |  gmap f (Full (na,la,pa,sa)) =
       let val x = f (sub (la,0))
           val n = ref (size la)
           val L = Array.array (!n,x)
           val _ = while (!n>0) do
                     (n := !n-1;Array.update (L,!n,f (sub (la,!n))))
        in
           Full (na,fromImpArray L,pa,sa)
       end

  (* predefined fold operations *)
  fun ufold f u (Empty _) = u
   |  ufold f u (Full (na,la,pred,suc)) =
      let val V = toImpArray na
          val n = Array.length V
          fun ufoldi x = 
	      if x<n then
                 let val c = (select p1 (fn (v,_)=>Array.sub (V,v)>0) 
			       (sub (pred,x)),x,sub (la,x),
                              select p1 (fn (v,i)=>Array.sub (V,v)>0) 
			       (sub (suc,x)))
		     val _ = Array.update (V,x,~1)
		     val r = ufoldi (x+1)
		  in f (c,r) end
	       else
		  u
	in ufoldi 0 end
      
  fun gfold f d b u l (Empty _) = u
   |  gfold f d b u l (Full (na,la,pred,suc)) =
      let val V = toImpArray na
          fun gfold1 v = (Array.update (V,v,~1);
                          let val l=sub (la,v)
                           in 
                              d (l,gfoldn (f (sub (pred,v),v,l,sub (suc,v))))
                          end)
          and gfoldn []         = u
           |  gfoldn ((v,i)::l) = 
              let val j = Array.sub (V,v)
               in
                  if j<0 orelse i<>j then gfoldn l
                                     else b (gfold1 v,gfoldn l)
              end
          and gfoldm []     = u
           |  gfoldm (v::l) = if Array.sub (V,v)<0 then gfoldm l
                                                   else b (gfold1 v,gfoldm l)
        in
           gfoldm l
       end

  fun mfold d b u l (Empty _) = u
   |  mfold d b u l (Full (na,la,pred,suc)) =
      let val V = toImpArray na
	  val n = Array.length V
	  val U = Array.array (n,false)
          val L = Array.array (n,NONE)
	  exception NotFound
	  fun e2word (v,w) = Word.fromInt (v*5+w)
	  val E = HashTable.mkTable (e2word,fn (v,w)=>v=w) (5*n+1,NotFound)

          fun mfold1 (pred,v) = 
	      (case pred of SOME z => HashTable.insert E ((z,v),true)
	        	  | NONE   => ();
	       if Array.sub (U,v) then 
                  forceOpt (Array.sub (L,v))
	       else 
	 	  (Array.update (U,v,true);
		   let val x=d (sub (la,v),mfoldn (v,sub (suc,v)))
		    in (Array.update (L,v,SOME x); x)
                   end))
          and mfoldn (_,[])       = u
           |  mfoldn (z,(v,i)::l) = 
              let val j = Array.sub (V,v)
		  val e = getOpt (HashTable.find E (z,v),false)
               in
                  if j<0 orelse i<>j orelse e then 
		     mfoldn (z,l)
                  else 
		     b (mfold1 (SOME z,v),mfoldn (z,l))
              end
          and mfoldm []     = u
           |  mfoldm (v::l) = 
	      if Array.sub (V,v)<0 then mfoldm l
                                   else b (mfold1 (NONE,v),mfoldm l)
        in
           mfoldm l
       end

  (* build a graph from a list of edges; x is a label used
     for all nodes *)
  fun el2gr x el = 
      let fun maxNode []         = 0
           |  maxNode ((v,w)::l) = Int.max (Int.max (v,w),maxNode l)
          val n = maxNode el+1
          val N = Array.array (n,0)
          val P = Array.array (n,[]:(node * stamp) list)
          val S = Array.array (n,[]:(node * stamp) list)
          fun scan []          = ()
           |  scan ((v,w)::el) = (Array.update (N,v,1);Array.update (N,w,1);
                                  Array.update (P,w,(v,1)::Array.sub (P,w));
                                  Array.update (S,v,(w,1)::Array.sub (S,v));
                                  scan el)
       in
          (scan el;
           Full (fromImpArray N,fromImpArray (Array.array (n,x)),
                 fromImpArray P,fromImpArray S))
      end

  (* build a graph from a list of node labels and edges *)
  fun mkgr nl el =
      let fun maxNode []         = 0
           |  maxNode ((v,w)::l) = Int.max (Int.max (v,w),maxNode l)
          val n = maxNode el+1
          val N = Array.array (n,0)
          val L = Array.fromList nl
          val P = Array.array (n,[]:(node * stamp) list)
          val S = Array.array (n,[]:(node * stamp) list)
          fun scan []          = ()
           |  scan ((v,w)::el) = (Array.update (N,v,1);Array.update (N,w,1);
                                  Array.update (P,w,(v,1)::Array.sub (P,w));
                                  Array.update (S,v,(w,1)::Array.sub (S,v));
                                  scan el)
       in
          (scan el;
           Full (fromImpArray N,fromImpArray L,fromImpArray P,fromImpArray S))
      end
end (* functor Graph *)
;
